// File author is Ítalo Lima Marconato Matias
//
// Created on May 11 of 2018, at 13:14 BRT
// Last edited on March 29 of 2019, at 22:26 BRT

#include <chicago/alloc.h>
#include <chicago/arch.h>
#include <chicago/console.h>
#include <chicago/debug.h>
#include <chicago/display.h>
#include <chicago/version.h>

#define TEST(name) static Int Test ## name(Void)																			// Macros for making and executing the tests
#define CALL_TEST(name) if (Test ## name()) { pass++; } else { fail++; } tests++;

TEST(Allocator0) {
	Int *arr = (Int*)MemAllocate(sizeof(Int));																				// Alloc some memory
	
	if (arr == Null) {
		return False;																										// Failed :(
	}
	
	*arr = 10;																												// Assign some value
	
	Int ret = *arr == 10;																									// Check if the value is ok
	
	MemFree((UIntPtr)arr);																									// Free the allocated memory
	
	return ret;
}

TEST(Allocator1) {
	UIntPtr old = 0;																										// Let's alloc a lot of small blocks!
	
	for (Int i = 0; i < 10000; i++) {
		UIntPtr addr = MemAllocate(sizeof(Int));																			// Alloc some memory
		
		if (addr == 0) {
			return False;																									// Failed :(
		}
		
		if (old != 0 && (addr > old)) {																						// Check if this one is in the same address of the old one
			MemFree(addr);																									// Isn't :(
			return False;
		} else if (old == 0) {																								// Set the 'old' variable if it isn't set
			old = addr;
		}
		
		MemFree(addr);																										// Free the allocated address
	}
	
	return True;
}

TEST(Allocator2) {
	Int arr[10000];																											// Let's test allocating one big block, a lot of small blocks, assigning some values, checking them and after all, free everything
	Int **arr2 = (Int**)MemAllocate(sizeof(Int*) * 10000);																	// Alloc some memory
	Int tmp = 0;
	
	if (arr2 == Null) {
		return False;																										// Failed
	}
	
	for (Int i = 0; i < 10000; i++) {																						// Let's allocate and set the values!
		tmp = (Int)ArchGetSeconds();																						// Get some value
		arr[i] = tmp;																										// Assign to our local array
		arr2[i] = (Int*)MemAllocate(sizeof(Int));																			// Alloc some memory
		
		if (arr2[i] == Null) {																								// Failed?
			for (Int j = 0; j < i; j++) {																					// Yes, free all the small block that we allocated
				MemFree((UIntPtr)arr2[j]);
			}
			
			MemFree((UIntPtr)arr2);																							// And the big block
			
			return False;
		}
		
		*(arr2[i]) = tmp;																									// Now, just assign the value!
	}
	
	for (Int i = 0; i < 10000; i++) {																						// Now let's check everything!
		if (*(arr2[i]) != arr[i]) {																							// Same value?
			for (Int j = i; j < 10000; j++) {																				// Nope, let's free everything :(
				MemFree((UIntPtr)arr2[j]);
			}
			
			MemFree((UIntPtr)arr2);
			
			return False;
		}
		
		MemFree((UIntPtr)arr2[i]);																							// Ok, so let's free this block!
	}
	
	MemFree((UIntPtr)arr2);																									// Free the big block
	
	return True;
}

TEST(Allocator3) {
	Int **arr = (Int**)MemAllocate(sizeof(Int*) * 3000000);																	// Alloc one very big block
	
	if (arr == Null) {
		return False;																										// Failed ;(
	}
	
	for (Int i = 0; i < 10000; i++) {																						// Now a lot of small block (just like the Allocator2 test)
		arr[i] = (Int*)MemAllocate(sizeof(Int));
		
		if (arr[i] == Null) {
			for (Int j = 0; j < i; j++) {																					// Failed, free everything :(
				MemFree((UIntPtr)arr[j]);
			}
			
			MemFree((UIntPtr)arr);
			
			return False;
		}
	}
	
	for (Int i = 0; i < 10000; i++) {																						// Now, let's free everything
		MemFree((UIntPtr)arr[i]);
	}
	
	MemFree((UIntPtr)arr);																									// Including our 'arr'
	
	UIntPtr new = MemAllocate(50000);																						// Allocate another big block
	MemFree(new);																											// And free it
	
	return new <= (UIntPtr)arr;																								// Return if the address of the new block was in the address of the old one
}

Void KernelMain(Void) {
	ArchInitDebug();																										// Init the architecture-dependent debugging method
	DbgWriteFormated("[Kernel] Arch debugging initialized\r\n");
	
	DbgWriteFormated("[Kernel] CHicago %s (codename %s, for %s)\r\n", CHICAGO_VSTR_C, CHICAGO_CODENAME_C, CHICAGO_ARCH_C);	// Print the system version
	
	ArchInitFPU();																											// Init the architecture-dependent FPU (floating point unit)
	DbgWriteFormated("[Kernel] Arch FPU initialized\r\n");
	
	ArchInitPMM();																											// Init the physical memory manager
	DbgWriteFormated("[Kernel] Arch PMM initialized\r\n");
	
	ArchInitVMM();																											// Init the virtual memory manager
	DbgWriteFormated("[Kernel] Arch VMM initialized\r\n");
	
	ArchInitDisplay();																										// Init the display
	DbgWriteFormated("[Kernel] Arch display initialized\r\n");
	
	ArchInit();																												// Let's finish it by initalizating the other architecture-dependent bits of the kernel
	DbgWriteFormated("[Kernel] Arch initialized\r\n");	
	
	ConClearScreen();																										// Clear the screen
	ConWriteFormated(L"CHicago Operating System for %s\r\n", CHICAGO_ARCH);													// Print some system informations
	ConWriteFormated(L"Codename '%s'\r\n", CHICAGO_CODENAME);
	ConWriteFormated(L"Version %d.%d.%d\r\n\r\n", CHICAGO_MAJOR, CHICAGO_MINOR, CHICAGO_BUILD);
	
	UIntPtr tests = 0;
	UIntPtr pass = 0;
	UIntPtr fail = 0;
	
	CALL_TEST(Allocator0);																									// Run all the allocator tests!
	CALL_TEST(Allocator1);
	CALL_TEST(Allocator2);
	CALL_TEST(Allocator3);
	
	ConWriteFormated(L"Number of tests: %d\r\nOK: %d\r\nFail: %d\r\n\r\n", tests, pass, fail);								// Now print the summary
	ArchHalt();																												// Halt
}
